#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import os
import numpy as np
import pandas as pd
import pytest

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    exp,
    Objective,
    SolverFactory,
    Suffix,
    Var,
)
import pyomo.contrib.parmest.parmest as parmest

from idaes.apps.uncertainty_propagation.uncertainties import (
    quantify_propagate_uncertainty,
    propagate_uncertainty,
    clean_variable_name,
)

ipopt_available = SolverFactory("ipopt").available(exception_flag=False)
kaug_available = SolverFactory("k_aug").available(exception_flag=False)
dotsens_available = SolverFactory("dot_sens").available(exception_flag=False)


@pytest.mark.skipif(not ipopt_available, reason="The 'ipopt' command is not available")
@pytest.mark.skipif(not kaug_available, reason="The 'k_aug' command is not available")
@pytest.mark.skipif(
    not dotsens_available, reason="The 'dot_sens' command is not available"
)
class TestUncertaintyPropagation:
    @pytest.mark.unit
    def test_quantify_propagate_uncertainty1(self):
        """
        It tests the function quantify_propagate_uncertainty with rooney & biegler's model.
        """
        from idaes.apps.uncertainty_propagation.examples.rooney_biegler import (
            rooney_biegler_model,
            rooney_biegler_model_opt,
        )

        variable_name = ["asymptote", "rate_constant"]
        data = pd.DataFrame(
            data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
            columns=["hour", "y"],
        )

        def SSE(model, data):
            expr = sum(
                (data.y[i] - model.response_function[data.hour[i]]) ** 2
                for i in data.index
            )
            return expr

        results = quantify_propagate_uncertainty(
            rooney_biegler_model, rooney_biegler_model_opt, data, variable_name, SSE
        )

        assert results.obj == pytest.approx(4.331711213656886, abs=1e-8, rel=1e-8)
        np.testing.assert_array_almost_equal(
            results.theta, [19.142575284617866, 0.53109137696521]
        )
        assert list(results.theta.keys()) == ["asymptote", "rate_constant"]
        np.testing.assert_array_almost_equal(results.gradient_f, [0.99506259, 0.945148])
        assert list(results.propagation_c) == []
        np.testing.assert_array_almost_equal(
            results.dsdp.toarray(), [[1.0, 0.0], [0.0, 1.0]]
        )
        np.testing.assert_array_almost_equal(
            results.cov, np.array([[6.30579403, -0.4395341], [-0.4395341, 0.04193591]])
        )
        assert results.propagation_f == pytest.approx(
            5.45439337747349, abs=1e-8, rel=1e-8
        )

    @pytest.mark.component
    def test_quantify_propagate_uncertainty2(self):
        """
        This is the same test as test_quantify_propagate_uncertainty1,
        but with the second argument of quantify_propagate_uncertainty as Pyomo Concrete Model.
        """
        from idaes.apps.uncertainty_propagation.examples.rooney_biegler import (
            rooney_biegler_model,
        )

        variable_name = ["asymptote", "rate_constant"]
        data = pd.DataFrame(
            data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
            columns=["hour", "y"],
        )

        def SSE(model, data):
            expr = sum(
                (data.y[i] - model.response_function[data.hour[i]]) ** 2
                for i in data.index
            )
            return expr

        model_uncertain = ConcreteModel()
        model_uncertain.asymptote = Var(initialize=15)
        model_uncertain.rate_constant = Var(initialize=0.5)
        model_uncertain.obj = Objective(
            expr=model_uncertain.asymptote
            * (1 - exp(-model_uncertain.rate_constant * 10)),
        )

        results = quantify_propagate_uncertainty(
            rooney_biegler_model, model_uncertain, data, variable_name, SSE
        )

        assert results.obj == pytest.approx(4.331711213656886, abs=1e-8, rel=1e-8)
        np.testing.assert_array_almost_equal(
            results.theta, [19.142575284617866, 0.53109137696521]
        )
        assert list(results.theta.keys()) == ["asymptote", "rate_constant"]
        np.testing.assert_array_almost_equal(results.gradient_f, [0.99506259, 0.945148])
        assert list(results.propagation_c) == []
        np.testing.assert_array_almost_equal(
            results.dsdp.toarray(), [[1.0, 0.0], [0.0, 1.0]]
        )
        np.testing.assert_array_almost_equal(
            results.cov, np.array([[6.30579403, -0.4395341], [-0.4395341, 0.04193591]])
        )
        assert results.propagation_f == pytest.approx(5.45439337747349)

    @pytest.mark.component
    def test_propagate_uncertainty(self):
        """
        It tests the function propagate_uncertainty with rooney & biegler's model.
        """
        from idaes.apps.uncertainty_propagation.examples.rooney_biegler import (
            rooney_biegler_model,
        )

        variable_name = ["asymptote", "rate_constant"]
        data = pd.DataFrame(
            data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
            columns=["hour", "y"],
        )

        def SSE(model, data):
            expr = sum(
                (data.y[i] - model.response_function[data.hour[i]]) ** 2
                for i in data.index
            )
            return expr

        parmest_class = parmest.Estimator(
            rooney_biegler_model, data, variable_name, SSE
        )
        obj, theta, cov = parmest_class.theta_est(calc_cov=True, cov_n=len(data.index))
        model_uncertain = ConcreteModel()
        model_uncertain.asymptote = Var(initialize=15)
        model_uncertain.rate_constant = Var(initialize=0.5)
        model_uncertain.obj = Objective(
            expr=model_uncertain.asymptote
            * (1 - exp(-model_uncertain.rate_constant * 10)),
        )

        propagate_results = propagate_uncertainty(
            model_uncertain, theta, cov, variable_name
        )

        np.testing.assert_array_almost_equal(
            propagate_results.gradient_f, [0.9950625870024135, 0.9451480001755206]
        )
        assert list(propagate_results.gradient_c) == []
        np.testing.assert_array_almost_equal(
            propagate_results.dsdp.toarray(), [[1.0, 0.0], [0.0, 1.0]]
        )
        assert list(propagate_results.propagation_c) == []
        assert propagate_results.propagation_f == pytest.approx(5.45439337747349)

    @pytest.mark.component
    def test_propagate_uncertainty1(self):
        """
        It tests the function propagate_uncertainty with
        min f:  p1*x1+ p2*(x2^2) + p1*p2
         s.t  c1: x1 + x2 = p1
              c2: x2 + x3 = p2
              0 <= x1, x2, x3 <= 10
              p1 = 10
              p2 = 5

        Variables = (x1, x2, x3)
        Parameters (fixed variables) = (p1, p2)
        """
        ### Create optimization model
        m = ConcreteModel()
        m.dual = Suffix(direction=Suffix.IMPORT)

        m.x1 = Var()
        m.x2 = Var()
        m.x3 = Var()

        # Define parameters
        m.p1 = Var(initialize=10)
        m.p2 = Var(initialize=5)
        m.p1.fix()
        m.p2.fix()

        # Define constraints
        m.con1 = Constraint(expr=m.x1 + m.x2 - m.p1 == 0)
        m.con2 = Constraint(expr=m.x2 + m.x3 - m.p2 == 0)

        # Define objective
        m.obj = Objective(expr=m.p1 * m.x1 + m.p2 * (m.x2**2) + m.p1 * m.p2)

        ### Solve optimization model
        opt = SolverFactory("ipopt", tee=True)
        opt.solve(m)

        ### Analytic solution
        """
        At the optimal solution, none of the bounds are active. As long as the active set
        does not change (i.e., none of the bounds become active), the
        first order optimality conditions reduce to a simple linear system.
        """

        # dual variables (multipliers)
        v2_ = 0
        v1_ = m.p1()

        # primal variables
        x2_ = (v1_ + v2_) / (2 * m.p2())
        x1_ = m.p1() - x2_
        x3_ = m.p2() - x2_

        ### Analytic sensitivity
        """
        Using the analytic solution above, we can compute the sensitivies of x and v to
        perturbations in p1 and p2.
        The matrix dx_dp constrains the sensitivities of x to perturbations in p
        """

        # Initialize sensitivity matrix Nx x Np
        # Rows: variables x
        # Columns: parameters p
        dx_dp = np.zeros((3, 2))
        # dx2/dp1 = 1/(2 * p2)
        dx_dp[1, 0] = 1 / (2 * m.p2())
        # dx2/dp2 = -(v1 + v2)/(2 * p2**2)
        dx_dp[1, 1] = -(v1_ + v2_) / (2 * m.p2() ** 2)
        # dx1/dp1 = 1 - dx2/dp1
        dx_dp[0, 0] = 1 - dx_dp[1, 0]
        # dx1/dp2 = 0 - dx2/dp2
        dx_dp[0, 1] = 0 - dx_dp[1, 1]
        # dx3/dp1 = 1 - dx2/dp1
        dx_dp[2, 0] = 0 - dx_dp[1, 0]
        # dx3/dp2 = 0 - dx2/dp2
        dx_dp[2, 1] = 1 - dx_dp[1, 1]

        """
        Similarly, we can compute the gradients df_dx, df_dp
        and Jacobians dc_dx, dc_dp
        """

        # Initialize 1 x 3 array to store (\partial f)/(\partial x)
        # Elements: variables x
        df_dx = np.zeros(3)
        # df/dx1 = p1
        df_dx[0] = m.p1()
        # df/dx2 = p2
        df_dx[1] = 2 * m.p2() * x2_
        # df/dx3 = 0
        # Initialize 1 x 2 array to store (\partial f)/(\partial p)
        # Elements: parameters p
        df_dp = np.zeros(2)
        # df/dxp1 = x1 + p2
        df_dp[0] = x1_ + m.p2()
        # df/dp2 = x2**2 + p1
        df_dp[1] = x2_**2 + m.p1()

        # Initialize 2 x 3 array to store (\partial c)/(\partial x)
        # Rows: constraints c
        # Columns: variables x
        dc_dx = np.zeros((2, 3))
        # dc1/dx1 = 1
        dc_dx[0, 0] = 1
        # dc1/dx2 = 1
        dc_dx[0, 1] = 1
        # dc2/dx2 = 1
        dc_dx[1, 1] = 1
        # dc2/dx3 = 1
        dc_dx[1, 2] = 1

        # Remaining entries are 0
        # Initialize 2 x 2 array to store (\partial c)/(\partial x)
        # Rows: constraints c
        # Columns: variables x
        dc_dp = np.zeros((2, 2))
        # dc1/dp1 = -1
        dc_dp[0, 0] = -1
        # dc2/dp2 = -1
        dc_dp[1, 1] = -1

        ### Uncertainty propagation
        """
        Now lets test the uncertainty propagation package. We will assume p has covariance
        sigma_p = [[2, 0], [0, 1]]
        """

        ## Prepare inputs
        # Covariance matrix
        sigma_p = np.array([[2, 0], [0, 1]])

        # Nominal values for uncertain parameters
        theta = {"p1": m.p1(), "p2": m.p2()}

        # Names of uncertain parameters
        theta_names = ["p1", "p2"]

        # Important to unfix the parameters!
        # Otherwise k_aug will complain about too few degrees of freedom
        m.p1.unfix()
        m.p2.unfix()

        ## Run package
        results = propagate_uncertainty(m, theta, sigma_p, theta_names)

        ## Check results

        tmp_f = df_dp + df_dx @ dx_dp
        sigma_f = tmp_f @ sigma_p @ tmp_f.transpose()

        tmp_c = dc_dp + dc_dx @ dx_dp
        sigma_c = tmp_c @ sigma_p @ tmp_c.transpose()

        # This currently just checks if the order of the outputs did not change
        # TODO: improve test robustness by using this information to set
        # var_idx and theta_idx. This way the test will still work
        # regardless of the order. In other words, the analytic solution needs to be
        # reordered to match the variable/constraint order from
        # this package. Alternately, the results could be converted into a Pandas dataframe

        # Build var_idx and theta_idx dynamically from results ordering
        # This should be robust to changes in ordering as long as the x's and p's remain in
        # the same relative order
        var_idx = np.array([False if c in ["p1", "p2"] else True for c in results.col])
        theta_idx = np.array(
            [True if c in ["p1", "p2"] else False for c in results.col]
        )

        # Check the gradient of the objective w.r.t. x matches
        np.testing.assert_array_almost_equal(
            results.gradient_f[var_idx], np.array(df_dx)
        )

        # Check the gradient of the objective w.r.t. p (parameters) matches
        np.testing.assert_array_almost_equal(
            results.gradient_f[theta_idx], np.array(df_dp)
        )

        # Check the Jacobian of the constraints w.r.t. x matches
        np.testing.assert_array_almost_equal(
            results.gradient_c.toarray()[:, var_idx], np.array(dc_dx)
        )

        # Check the Jacobian of the constraints w.r.t. p (parameters) matches
        np.testing.assert_array_almost_equal(
            results.gradient_c.toarray()[:, theta_idx], np.array(dc_dp)
        )

        # Check the NLP sensitivity results for the variables (x) matches
        np.testing.assert_array_almost_equal(
            results.dsdp.toarray()[var_idx, :], np.array(dx_dp)
        )

        # Check the NLP sensitivity results for the parameters (p) matches
        np.testing.assert_array_almost_equal(
            results.dsdp.toarray()[theta_idx, :], np.array([[1, 0], [0, 1]])
        )

        # Check the uncertainty propagation results for the constrains matches
        np.testing.assert_array_almost_equal(results.propagation_c, np.sum(sigma_c))

        # Check the uncertainty propagation results for the objective matches
        assert results.propagation_f == pytest.approx(sigma_f)

    @pytest.mark.component
    def test_propagate_uncertainty_error(self):
        """
        It tests a TypeError when the model_uncertain of function propagate_uncertainty is neither python function nor Pyomo ConcreteModel
        """
        from idaes.apps.uncertainty_propagation.examples.rooney_biegler import (
            rooney_biegler_model,
        )

        variable_name = ["asymptote", "rate_constant"]
        data = pd.DataFrame(
            data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
            columns=["hour", "y"],
        )

        def SSE(model, data):
            expr = sum(
                (data.y[i] - model.response_function[data.hour[i]]) ** 2
                for i in data.index
            )
            return expr

        parmest_class = parmest.Estimator(
            rooney_biegler_model, data, variable_name, SSE
        )
        obj, theta, cov = parmest_class.theta_est(calc_cov=True, cov_n=len(data.index))
        model_uncertain = ConcreteModel()
        model_uncertain.asymptote = Var(initialize=15)
        model_uncertain.rate_constant = Var(initialize=0.5)
        model_uncertain.obj = Objective(
            expr=model_uncertain.asymptote
            * (1 - exp(-model_uncertain.rate_constant * 10)),
        )
        with pytest.raises(TypeError):
            propagate_results = propagate_uncertainty(1, theta, cov, variable_name)

    @pytest.mark.integration
    def test_quantify_propagate_uncertainty_NRTL(self):
        """
        It tests the function quantify_propagate_uncertainty with IDAES NRTL model.
        """
        from idaes.apps.uncertainty_propagation.examples.NRTL_model_scripts import (
            NRTL_model,
            NRTL_model_opt,
        )

        variable_name = [
            "fs.properties.tau['benzene','toluene']",
            "fs.properties.tau['toluene','benzene']",
        ]
        current_path = os.path.dirname(os.path.realpath(__file__))
        data = pd.read_csv(os.path.join(current_path, "BT_NRTL_dataset.csv"))

        def SSE(model, data):
            expr = (
                float(data["vap_benzene"])
                - model.fs.flash.vap_outlet.mole_frac_comp[0, "benzene"]
            ) ** 2 + (
                float(data["liq_benzene"])
                - model.fs.flash.liq_outlet.mole_frac_comp[0, "benzene"]
            ) ** 2
            return expr * 1e4

        results = quantify_propagate_uncertainty(
            NRTL_model, NRTL_model_opt, data, variable_name, SSE
        )
        np.testing.assert_array_almost_equal(results.obj, 5.074968578304798)
        np.testing.assert_array_almost_equal(
            np.fromiter(results.theta.values(), dtype=float), [-0.8987624, 1.41048611]
        )
        np.testing.assert_array_almost_equal(results.gradient_f[0], [-0.19649493])
        np.testing.assert_almost_equal(
            results.cov,
            np.array([[0.01194738, -0.02557055], [-0.02557055, 0.05490639]]),
        )
        assert results.propagation_f == pytest.approx(0.0021199499778127204)

    @pytest.mark.integration
    @pytest.mark.xfail
    # TODO: This test fails as the problem being tested is in fact feasible
    # This should be replaced with unit testing of the underlying methods which
    # can test the expected fail in more controlled circumstances (and faster)
    def test_quantify_propagate_uncertainty_NRTL_exception(self):
        """
        It tests an exception error when the ipopt fails for the function quantify_propagate_uncertainty with IDAES NRTL model.
        """
        from idaes.apps.uncertainty_propagation.examples.NRTL_model_scripts import (
            NRTL_model,
            NRTL_model_opt_infeasible,
        )

        variable_name = [
            "fs.properties.tau['benzene','toluene']",
            "fs.properties.tau['toluene','benzene']",
        ]
        current_path = os.path.dirname(os.path.realpath(__file__))
        data = pd.read_csv(os.path.join(current_path, "BT_NRTL_dataset.csv"))

        def SSE(model, data):
            expr = (
                float(data["vap_benzene"])
                - model.fs.flash.vap_outlet.mole_frac_comp[0, "benzene"]
            ) ** 2 + (
                float(data["liq_benzene"])
                - model.fs.flash.liq_outlet.mole_frac_comp[0, "benzene"]
            ) ** 2
            return expr * 1e4

        with pytest.raises(RuntimeError):
            results = quantify_propagate_uncertainty(
                NRTL_model,
                NRTL_model_opt_infeasible,
                data,
                variable_name,
                SSE,
            )

    @pytest.mark.unit
    def test_Exception1(self):
        """
        It tests an ValueError when the tee is not bool for the function quantify_propagate_uncertainty with rooney & biegler's model.
        """
        from idaes.apps.uncertainty_propagation.examples.rooney_biegler import (
            rooney_biegler_model,
            rooney_biegler_model_opt,
        )

        variable_name = ["asymptote", "rate_constant"]
        data = pd.DataFrame(
            data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
            columns=["hour", "y"],
        )

        def SSE(model, data):
            expr = sum(
                (data.y[i] - model.response_function[data.hour[i]]) ** 2
                for i in data.index
            )
            return expr

        tee = 1
        with pytest.raises(TypeError):
            results = quantify_propagate_uncertainty(
                rooney_biegler_model,
                rooney_biegler_model_opt,
                data,
                variable_name,
                SSE,
                tee,
            )

    @pytest.mark.unit
    def test_Exception2(self):
        """
        It tests an ValueError when the diagnostic_mode is not bool for the function quantify_propagate_uncertainty with rooney & biegler's model.
        """
        from idaes.apps.uncertainty_propagation.examples.rooney_biegler import (
            rooney_biegler_model,
            rooney_biegler_model_opt,
        )

        variable_name = ["asymptote", "rate_constant"]
        data = pd.DataFrame(
            data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
            columns=["hour", "y"],
        )

        def SSE(model, data):
            expr = sum(
                (data.y[i] - model.response_function[data.hour[i]]) ** 2
                for i in data.index
            )
            return expr

        tee = False
        diagnostic_mode = 1
        with pytest.raises(TypeError):
            results = quantify_propagate_uncertainty(
                rooney_biegler_model,
                rooney_biegler_model_opt,
                data,
                variable_name,
                SSE,
                tee,
                diagnostic_mode,
            )

    @pytest.mark.unit
    def test_Exception3(self):
        """
        It tests an ValeError when solver_options is not a dictionary for the function quantify_propagate_uncertainty with rooney & biegler's model.
        """
        from idaes.apps.uncertainty_propagation.examples.rooney_biegler import (
            rooney_biegler_model,
            rooney_biegler_model_opt,
        )

        variable_name = ["asymptote", "rate_constant"]
        data = pd.DataFrame(
            data=[[1, 8.3], [2, 10.3], [3, 19.0], [4, 16.0], [5, 15.6], [7, 19.8]],
            columns=["hour", "y"],
        )

        def SSE(model, data):
            expr = sum(
                (data.y[i] - model.response_function[data.hour[i]]) ** 2
                for i in data.index
            )
            return expr

        tee = False
        diagnostic_mode = False
        solver_options = [1e-8]
        with pytest.raises(TypeError):
            results = quantify_propagate_uncertainty(
                rooney_biegler_model,
                rooney_biegler_model_opt,
                data,
                variable_name,
                SSE,
                tee,
                diagnostic_mode,
                solver_options,
            )

    @pytest.mark.unit
    def test_clean_variable_name1(self):
        """
        It tests the function clean_variable_name when variable names contain ' and spaces.
        """
        theta_names = [
            "fs.properties.tau['benzene', 'toluene']",
            "fs.properties.tau['toluene', 'benzene' ]",
        ]
        theta_names_new, var_dic, clean = clean_variable_name(theta_names)
        theta_names_expected = [
            "fs.properties.tau[benzene,toluene]",
            "fs.properties.tau[toluene,benzene]",
        ]
        assert len(theta_names_expected) == len(theta_names_new)
        assert all([a == b for a, b in zip(theta_names_expected, theta_names_new)])

        assert len(theta_names_expected) == len(var_dic.keys())
        assert all(
            [
                a == b
                for a, b in zip(sorted(theta_names_expected), sorted(var_dic.keys()))
            ]
        )

        assert len(theta_names) == len(var_dic.values())
        assert all(
            [a == b for a, b in zip(sorted(theta_names), sorted(var_dic.values()))]
        )
        assert clean == True

    @pytest.mark.unit
    def test_clean_variable_name2(self):
        """
        It tests the function clean_variable_name when variable names do not contain any ' and spaces.
        """
        theta_names = [
            "fs.properties.tau[benzene,toluene]",
            "fs.properties.tau[toluene,benzene]",
        ]
        theta_names_new, var_dic, clean = clean_variable_name(theta_names)
        assert len(theta_names) == len(theta_names_new)
        assert all([a == b for a, b in zip(theta_names, theta_names_new)])

        assert len(theta_names) == len(var_dic.keys())
        assert all(
            [a == b for a, b in zip(sorted(theta_names), sorted(var_dic.keys()))]
        )

        assert len(theta_names) == len(var_dic.values())
        assert all(
            [a == b for a, b in zip(sorted(theta_names), sorted(var_dic.values()))]
        )
        assert clean == False



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains phase equilibria utility functions for use in IDAES models.
"""

__author__ = "Alejandro Garciadiego"

import pytest

import numpy as np

import idaes.logger as idaeslog

from pyomo.environ import ConcreteModel

# Import Pyomo units
from pyomo.environ import units as pyunits
from pyomo.util.check_units import assert_units_equivalent

from idaes.core import LiquidPhase, VaporPhase, Component, PhaseType as PT

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil.smooth_VLE import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    IdealBubbleDew,
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import (
    fugacity,
    log_fugacity,
)

import idaes.models.properties.modular_properties.pure.NIST as NIST
import idaes.models.properties.modular_properties.pure.RPP5 as RPP5

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.core.util.phase_equilibria import TXYDataClass, Txy_data
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available


@pytest.mark.unit
def test_Txy_dataclass():
    TD = TXYDataClass(
        "benzene", "toluene", pyunits.kg / pyunits.m / pyunits.s**2, pyunits.K, 101325
    )
    TD.TBubb = [353.3205, 365.3478, 383.8817]
    TD.TDew = [353.3237, 372.0203, 383.8845]
    TD.x = [0.9999, 0.5, 0.01]

    assert TD.Component_1 == "benzene"
    assert TD.Component_2 == "toluene"
    assert_units_equivalent(TD.Punits, pyunits.kg / pyunits.m / pyunits.s**2)
    assert_units_equivalent(TD.Tunits, pyunits.K)
    assert TD.P == 101325
    assert TD.TBubb == [
        pytest.approx(353.3205, abs=1e-4),
        pytest.approx(365.3478, abs=1e-4),
        pytest.approx(383.8817, abs=1e-4),
    ]
    assert TD.TDew == [
        pytest.approx(353.3237, abs=1e-4),
        pytest.approx(372.0203, abs=1e-4),
        pytest.approx(383.8845, abs=1e-4),
    ]
    assert TD.x == [
        pytest.approx(0.9999, abs=1e-4),
        pytest.approx(0.5, abs=1e-4),
        pytest.approx(0.01, abs=1e-4),
    ]


# Author: Alejandro Garciadiego
@pytest.mark.component
def test_Txy_data():
    configuration = {
        # Specifying components
        "components": {
            "benzene": {
                "type": Component,
                "pressure_sat_comp": NIST,
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                "parameter_data": {
                    "mw": (78.1136e-3, pyunits.kg / pyunits.mol),  # [1]
                    "pressure_crit": (48.9e5, pyunits.Pa),  # [1]
                    "temperature_crit": (562.2, pyunits.K),  # [1]
                    "pressure_sat_comp_coeff": {
                        "A": (4.72583, None),  # [2]
                        "B": (1660.652, pyunits.K),
                        "C": (-1.461, pyunits.K),
                    },
                },
            },
            "toluene": {
                "type": Component,
                "pressure_sat_comp": NIST,
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                "parameter_data": {
                    "mw": (92.1405e-3, pyunits.kg / pyunits.mol),  # [1]
                    "pressure_crit": (41e5, pyunits.Pa),  # [1]
                    "temperature_crit": (591.8, pyunits.K),  # [1]
                    "pressure_sat_comp_coeff": {
                        "A": (4.07827, None),  # [2]
                        "B": (1343.943, pyunits.K),
                        "C": (-53.773, pyunits.K),
                    },
                },
            },
        },
        # Specifying phases
        "phases": {
            "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
            "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
        },
        # Set base units of measurement
        "base_units": {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        # Specifying state definition
        "state_definition": FTPx,
        "state_bounds": {
            "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
            "temperature": (273.15, 300, 450, pyunits.K),
            "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
        },
        "pressure_ref": (1e5, pyunits.Pa),
        "temperature_ref": (300, pyunits.K),
        # Defining phase equilibria
        "phases_in_equilibrium": [("Vap", "Liq")],
        "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
        "bubble_dew_method": IdealBubbleDew,
    }

    model = ConcreteModel()

    model.params = GenericParameterBlock(**configuration)

    TD = Txy_data(
        model,
        "benzene",
        "toluene",
        101325,
        num_points=3,
        temperature=298.15,
        print_level=idaeslog.CRITICAL,
        solver_op={"tol": 1e-6},
    )

    assert TD.Component_1 == "benzene"
    assert TD.Component_2 == "toluene"
    assert_units_equivalent(TD.Punits, pyunits.kg / pyunits.m / pyunits.s**2)
    assert_units_equivalent(TD.Tunits, pyunits.K)
    assert TD.P == 101325
    assert TD.TBubb == [
        pytest.approx(353.4853, abs=1e-4),
        pytest.approx(365.2127, abs=1e-4),
        pytest.approx(383.2909, abs=1e-4),
    ]
    assert TD.TDew == [
        pytest.approx(353.7978, abs=1e-2),
        pytest.approx(371.8702, abs=1e-4),
        pytest.approx(383.5685, abs=1e-4),
    ]
    assert TD.x == [
        pytest.approx(0.99, abs=1e-4),
        pytest.approx(0.5, abs=1e-4),
        pytest.approx(0.01, abs=1e-4),
    ]


# Author: Alejandro Garciadiego
@pytest.mark.component
@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
def test_Txy_data_no_dew():
    configuration = {
        # Specifying components
        "components": {
            "bmimPF6": {
                "type": Component,
                "valid_phase_types": PT.liquidPhase,
                "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                "parameter_data": {
                    "mw": (284.18e-3, pyunits.kg / pyunits.mol),  # [1]
                    "pressure_crit": (24e5, pyunits.Pa),  # [1]
                    "temperature_crit": (860, pyunits.K),  # [1]
                    "omega": 0.7917,
                },
            },
            "carbon_dioxide": {
                "type": Component,
                "pressure_sat_comp": NIST,
                "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                "parameter_data": {
                    "mw": (44.010e-3, pyunits.kg / pyunits.mol),
                    "pressure_crit": (71.8e5, pyunits.Pa),
                    "temperature_crit": (304.1, pyunits.K),
                    "omega": 0.239,
                    "pressure_sat_comp_coeff": {
                        "A": (6.81228, None),
                        "B": (1301.679, pyunits.K),
                        "C": (-3.494, pyunits.K),
                    },
                },
            },
        },
        # Specifying phases
        "phases": {
            "Liq": {
                "type": LiquidPhase,
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
            "Vap": {
                "type": VaporPhase,
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
        },
        # Set base units of measurement
        "base_units": {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        # Specifying state definition
        "state_definition": FTPx,
        "state_bounds": {
            "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
            "temperature": (10, 300, 500, pyunits.K),
            "pressure": (5e-4, 1e5, 1e10, pyunits.Pa),
        },
        "pressure_ref": (101325, pyunits.Pa),
        "temperature_ref": (298.15, pyunits.K),
        # Defining phase equilibria
        "phases_in_equilibrium": [("Vap", "Liq")],
        "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
        "bubble_dew_method": LogBubbleDew,
        "parameter_data": {
            "PR_kappa": {
                ("bmimPF6", "bmimPF6"): 0.000,
                ("bmimPF6", "carbon_dioxide"): -0.05718,
                ("carbon_dioxide", "carbon_dioxide"): 0.000,
                ("carbon_dioxide", "bmimPF6"): 0.0293,
            }
        },
    }

    model = ConcreteModel()

    model.params = GenericParameterBlock(**configuration)

    # Again, add more calculation points because the PR for single-phase
    # component VLE initialization hasn't been accepted
    TD = Txy_data(
        model,
        "carbon_dioxide",
        "bmimPF6",
        101325,
        num_points=9,
        temperature=150.15,
        print_level=idaeslog.CRITICAL,
        solver_op={"tol": 1e-6},
    )

    assert TD.Component_1 == "carbon_dioxide"
    assert TD.Component_2 == "bmimPF6"
    assert_units_equivalent(TD.Punits, pyunits.kg / pyunits.m / pyunits.s**2)
    assert_units_equivalent(TD.Tunits, pyunits.K)
    assert TD.P == 101325
    TBubb_ref = np.array(
        [
            185.6468,
            183.9464,
            185.2789,
            187.8233,
            191.4697,
            196.6568,
            204.6795,
            220.183,
            331.1625,
        ]
    )
    assert TD.TBubb == pytest.approx(TBubb_ref, abs=1e-2)
    assert TD.TDew == []
    x_ref = np.array([0.99, 0.8675, 0.745, 0.6225, 0.5, 0.3775, 0.255, 0.1325, 0.01])
    assert TD.x == pytest.approx(x_ref, abs=1e-4)


# Author: Alejandro Garciadiego
@pytest.mark.component
@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
def test_Txy_data_no_bub():
    configuration = {
        # Specifying components
        "components": {
            "methane": {
                "type": Component,
                "valid_phase_types": PT.vaporPhase,
                "parameter_data": {
                    "mw": (16.043e-3, pyunits.kg / pyunits.mol),  # [1]
                    "pressure_crit": (46e5, pyunits.Pa),  # [1]
                    "temperature_crit": (190.4, pyunits.K),  # [1]
                    "omega": 0.011,
                },
            },
            "ethane": {
                "type": Component,
                "pressure_sat_comp": RPP5,
                "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                "parameter_data": {
                    "mw": (30.070e-3, pyunits.kg / pyunits.mol),  # [1]
                    "pressure_crit": (48.8e5, pyunits.Pa),  # [1]
                    "temperature_crit": (305.4, pyunits.K),  # [1]
                    "omega": 0.099,
                    "pressure_sat_comp_coeff": {
                        "A": (3.95405, None),  # [4]
                        "B": (663.720, pyunits.K),
                        "C": (256.681, pyunits.K),
                    },
                },
            },
        },
        # Specifying phases
        "phases": {
            "Liq": {
                "type": LiquidPhase,
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
            "Vap": {
                "type": VaporPhase,
                "equation_of_state": Cubic,
                "equation_of_state_options": {"type": CubicType.PR},
            },
        },
        # Set base units of measurement
        "base_units": {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        # Specifying state definition
        "state_definition": FTPx,
        "state_bounds": {
            "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
            "temperature": (10, 300, 1500, pyunits.K),
            "pressure": (5e-4, 1e5, 1e10, pyunits.Pa),
        },
        "pressure_ref": (101325, pyunits.Pa),
        "temperature_ref": (298.15, pyunits.K),
        # Defining phase equilibria
        "phases_in_equilibrium": [("Vap", "Liq")],
        "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
        "bubble_dew_method": LogBubbleDew,
        "parameter_data": {
            "PR_kappa": {
                ("methane", "methane"): 0.000,
                ("methane", "ethane"): 0.000,
                ("ethane", "ethane"): 0.000,
                ("ethane", "methane"): 0.000,
            }
        },
    }

    model = ConcreteModel()

    model.params = GenericParameterBlock(**configuration)

    # TODO: There probably should be a config option to run initialization at
    # each x or to use the initialization from the previous x. In the meantime,
    # add more points to avoid convergence issues. Besides, better
    # initialization for noncondensables is still languishing
    TD = Txy_data(
        model,
        "methane",
        "ethane",
        101325,
        num_points=9,
        temperature=298,
        print_level=idaeslog.CRITICAL,
        solver_op={"tol": 1e-6},
    )

    assert TD.Component_1 == "methane"
    assert TD.Component_2 == "ethane"
    assert_units_equivalent(TD.Punits, pyunits.kg / pyunits.m / pyunits.s**2)
    assert_units_equivalent(TD.Tunits, pyunits.K)
    assert TD.P == 101325
    assert TD.TBubb == []
    TDew_ref = np.array(
        [
            126.9026,
            153.1591,
            161.8914,
            167.6907,
            172.1489,
            175.8221,
            178.9746,
            181.7545,
            184.2535,
        ]
    )
    assert TD.TDew == pytest.approx(TDew_ref, abs=1e-2)
    x_ref = np.array([0.99, 0.8675, 0.745, 0.6225, 0.5, 0.3775, 0.255, 0.1325, 0.01])
    assert TD.x == pytest.approx(x_ref, abs=1e-4)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Rooney Biegler model, based on Rooney, W. C. and Biegler, L. T. (2001). Design for 
model parameter uncertainty using nonlinear confidence regions. AIChE Journal, 
47(8), 1794-1804.
"""
import pandas as pd
import pyomo.environ as pyo


def rooney_biegler_model(data):
    """This function generates an instance of the rooney & biegler Pyomo model using 'data' as the input argument

    Parameters
    ----------
    data: pandas DataFrame, list of dictionaries, or list of json file names
        Data that is used to build an instance of the Pyomo model

    Returns
    -------
    m: an instance of the Pyomo model
        for estimating parameters and covariance
    """
    model = pyo.ConcreteModel()

    model.asymptote = pyo.Var(initialize=15)
    model.rate_constant = pyo.Var(initialize=0.5)

    def response_rule(m, h):
        expr = m.asymptote * (1 - pyo.exp(-m.rate_constant * h))
        return expr

    model.response_function = pyo.Expression(data.hour, rule=response_rule)

    return model


def rooney_biegler_model_opt():
    """This function generates an instance of the rooney & biegler Pyomo model

    Returns
    -------
    m: an instance of the Pyomo model
        for uncertainty propagation
    """

    model = pyo.ConcreteModel()

    model.asymptote = pyo.Var(initialize=15)
    model.rate_constant = pyo.Var(initialize=0.5)

    model.obj = pyo.Objective(
        expr=model.asymptote * (1 - pyo.exp(-model.rate_constant * 10)),
        sense=pyo.minimize,
    )
    return model



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module is a collection of classes that provide a
friendlier interface to MPI (through mpi4py). They help
allocate local tasks/data from global tasks/data and gather
global data (from all processors).

Although general, this module was only implemented to
work with the convergence evaluation framework. More work
is needed to make this appropriate for general use.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from collections import OrderedDict
import importlib


class MPIInterface:
    __have_mpi__ = None

    def __init__(self):
        if MPIInterface.__have_mpi__ is None:
            # This is trying to import mpy4py.MPI, and setting a flag to indicate
            # if it succeeds or not.
            # we do this here instead of at the module level, because we only want
            # to do the import if an MPIInterface is ever requested.
            try:
                # try the import (the 'globals()' makes sure it is imported
                # in the module space and not local to the __init__ method)
                globals()["MPI"] = importlib.import_module("mpi4py.MPI")
                # import succeeded
                MPIInterface.__have_mpi__ = True
            except ImportError:
                # import failed (e.g., no mpi4py installed)
                MPIInterface.__have_mpi__ = False

        self._comm = None
        self._size = None
        self._rank = None

        if self.have_mpi:
            self._comm = MPI.COMM_WORLD  # pylint: disable=undefined-variable
            self._size = self._comm.Get_size()
            self._rank = self._comm.Get_rank()

    @property
    def have_mpi(self):
        assert MPIInterface.__have_mpi__ is not None
        return MPIInterface.__have_mpi__

    @property
    def comm(self):
        return self._comm

    @property
    def rank(self):
        return self._rank

    @property
    def size(self):
        return self._size


class ParallelTaskManager:
    def __init__(self, n_total_tasks, mpi_interface=None):
        if mpi_interface is None:
            self._mpi_interface = MPIInterface()
        else:
            self._mpi_interface = mpi_interface
        self._n_total_tasks = n_total_tasks

        if not self._mpi_interface.have_mpi:
            self._local_map = range(n_total_tasks)
        else:
            size = self._mpi_interface.size

            # there must be a better way to do this
            # find which entries in global correspond
            # to this process (want them to be contiguous
            # for the MPI Allgather calls later
            local_N = [0 for i in range(self._mpi_interface.size)]
            for i in range(n_total_tasks):
                process_i = i % size
                local_N[process_i] += 1

            start = 0
            end = None
            for i, v in enumerate(local_N):
                if i == self._mpi_interface.rank:
                    end = start + v
                    break
                else:
                    start += v

            self._local_map = list(range(start, end))

    def is_root(self):
        if not self._mpi_interface.have_mpi or self._mpi_interface.rank == 0:
            return True
        return False

    # ToDo: fix the parallel task manager to handle dictionaries as well as lists
    def global_to_local_data(self, global_data):
        if isinstance(global_data, list):
            local_data = list()
            assert len(global_data) == self._n_total_tasks
            for i in self._local_map:
                local_data.append(global_data[i])
            return local_data
        elif isinstance(global_data, OrderedDict):
            local_data = OrderedDict()
            assert len(global_data) == self._n_total_tasks
            idx = 0
            for k, v in global_data.items():
                if idx in self._local_map:
                    local_data[k] = v
                idx += idx
            return local_data
        raise ValueError(
            "Unknown type passed to global_to_local_data. Expected list or OrderedDict."
        )

    def allgather_global_data(self, local_data):
        assert len(local_data) == len(self._local_map)
        if not self._mpi_interface.have_mpi:
            return list(local_data)

        # PYLINT-TODO-FIX fix the error due to the
        # non-existing global_data_list_of_lists variable
        # pylint: disable=undefined-variable
        return self._stack_global_data(global_data_list_of_lists)

    def gather_global_data(self, local_data):
        assert len(local_data) == len(self._local_map)
        if not self._mpi_interface.have_mpi:
            return list(local_data)

        comm = self._mpi_interface.comm
        global_data_list_of_lists = comm.gather(local_data)

        if global_data_list_of_lists is not None:
            return self._stack_global_data(global_data_list_of_lists)

        assert self.is_root() is False
        return None

    def _stack_global_data(self, global_data_list_of_lists):
        # stack the list of lists into one global data list
        # ToDo: test that this is equivalent to [d for sublist in global_data_list_of_lists for d in sublist]
        global_data = list()
        for i in range(self._mpi_interface.size):
            global_data.extend(global_data_list_of_lists[i])
        return global_data
